#!/usr/bin/python3
#
# Author: Markus Fuchs (fuchs@sero-systems.de)
# Created: 2019-08-14
#
#

import functools
import inspect
import logging
import socket
import struct
import sys

import dbus
from sensor_web_ui.network_manager import NM_BUS_NAME, nm_device_type, device_states, nm_state, nm_80211_mode, \
    DEFAULT_WIFI_HOTSPOT_SETTINGS, DEFAULT_ETHERNET_SETTINGS, conn_nm

logger = logging.getLogger("network")
bus, settings, manager = conn_nm()


def catch_dbus_conn_err(f):
    """Decorator to catch DBUS service errors and re-connect to NetworkManager after it was restarted"""
    global bus, settings, manager

    @functools.wraps(f)
    def func(*args, **kwargs):
        global bus, settings, manager
        try:
            return f(*args, **kwargs)
        except dbus.exceptions.DBusException as e:
            if e.get_dbus_name() == "org.freedesktop.DBus.Error.ServiceUnknown":
                bus, settings, manager = conn_nm()
                return f(*args, **kwargs)

    return func


def decorate_all_methods(decorator):
    """Class decorator which decorates all methods of a class"""
    def apply_decorator(cls):
        for k, f in cls.__dict__.items():
            if inspect.isfunction(f):
                setattr(cls, k, decorator(f))
        return cls
    return apply_decorator


def identity(x):
    return x


def conv_addr(av):
    """Convert IP settings address prefix to DBUS uint32"""
    av["prefix"] = dbus.UInt32(av["prefix"])
    return av


def ip_to_int(dns):
    """Convert IP string to int in network byte order"""
    if type(dns) == str:
        return struct.unpack("!I", socket.inet_aton(dns))[0]
    return dns


def ip_to_bytearray(dns):
    if type(dns) == str:
        return socket.inet_pton(socket.AF_INET6, dns)
    return dns


SETTINGS_TRANSFORMER = {
    "connection": {
        "type": identity,
        "id": identity,
        "uuid": identity
    },
    "802-3-ethernet": {
        "duplex": identity
    },
    "802-11-wireless": {
        "ssid": lambda x: dbus.ByteArray(x.encode("utf-8")),
        "mode": identity,
        "band": identity,
        "channel": lambda x: dbus.UInt32(x)
    },
    "ipv4": {
        "address-data": lambda xl: dbus.Array([conv_addr(av) for av in xl], signature=dbus.Signature("a{sv}")),
        "gateway": identity,
        "method": identity,
        "dns": lambda xl: dbus.Array([ip_to_int(d) for d in xl], signature=dbus.Signature("u"))
    },
    "ipv6": {
        "address-data": lambda xl: dbus.Array([conv_addr(av) for av in xl], signature=dbus.Signature("a{sv}")),
        "gateway": identity,
        "method": identity,
        "dns": lambda xl: dbus.Array([dbus.ByteArray(ip_to_bytearray(d)) for d in xl], signature=dbus.Signature("ay"))
    },
    "802-11-wireless-security": {
        "key-mgmt": identity,
        "psk": identity
    }
}


def transform_settings(s, trans=SETTINGS_TRANSFORMER):
    """Recursive function to map types of a settings dict to types expected by DBUS.
    Ignores fields not stated in trans."""
    res = {}
    for k, v in s.items():
        if k not in trans.keys():
            continue
        transformer = trans[k]

        if callable(transformer):
            res[k] = transformer(v)
        elif type(transformer) == dict and type(v) == dict:
            tmp = transform_settings(v, transformer)
            if tmp is not None:
                res[k] = tmp

    return res if len(res) > 0 else None


@catch_dbus_conn_err
def _get_properties(bus_name, path, clazz, properties):
    """Get properties from a DBUS object"""
    tmp_proxy = bus.get_object(bus_name, path)
    prop_iface = dbus.Interface(tmp_proxy, "org.freedesktop.DBus.Properties")

    res = {}
    for p in properties:
        try:
            r = prop_iface.Get(clazz, p["name"])

            if isinstance(r, dbus.ObjectPath) and p["sub"] is not None:
                r = _get_properties(bus_name, r, p["sub"]["class"], p["sub"]["props"])

            res[p["name"]] = p["map"](r)
        except dbus.exceptions.DBusException:
            res[p["name"]] = None

    return res


def get_device_by_type(dev_type):
    infos = {"class": "org.freedesktop.NetworkManager.Device", "props": [
            {"name": "Interface", "map": identity, "sub": None},
            {"name": "DeviceType", "map": lambda x: nm_device_type[x], "sub": None}
            ]
    }

    devs = [_get_properties(NM_BUS_NAME, dev_path, infos["class"], infos["props"])
            for dev_path in manager.GetDevices()]

    return [d for d in devs if d["DeviceType"] == dev_type]


@catch_dbus_conn_err
def get_device_state(interface):
    dev_path = manager.GetDeviceByIpIface(interface)

    infos = {"class": "org.freedesktop.NetworkManager.Device", "props": [
            {"name": "Interface", "map": identity, "sub": None},
            {"name": "DeviceType", "map": lambda x: nm_device_type[x], "sub": None},
            {"name": "Ip4Address", "map": identity, "sub": None},
            {"name": "Ip6Address", "map": identity, "sub": None},
            {"name": "State", "map": lambda x: device_states[x], "sub": None},
            {"name": "Ip4Config", "map": identity, "sub": {
                "class": "org.freedesktop.NetworkManager.IP4Config",
                "props": [
                    {"name": "AddressData", "map": identity, "sub": None},
                    {"name": "RouteData", "map": identity, "sub": None},
                    {"name": "Gateway", "map": identity, "sub": None},
                    {"name": "NameserverData", "map": identity, "sub": None},
                    {"name": "Domains", "map": identity, "sub": None}
                    ]
                }
            },
            {"name": "Ip6Config", "map": identity, "sub": {
                "class": "org.freedesktop.NetworkManager.IP6Config",
                "props": [
                    {"name": "AddressData", "map": identity, "sub": None},
                    {"name": "RouteData", "map": identity, "sub": None},
                    {"name": "Gateway", "map": identity, "sub": None},
                    {"name": "NameserverData", "map": identity, "sub": None},
                    {"name": "Domains", "map": identity, "sub": None}
                    ]
                }
            },
            # Get path for Settings.Connection object of active connection
            {"name": "ActiveConnection", "map": identity, "sub": {
                "class": "org.freedesktop.NetworkManager.Connection.Active",
                "props": [
                    {"name": "Connection", "map": identity, "sub": None}
                    ]
                }
            }
            ]
    }

    res = _get_properties(NM_BUS_NAME, dev_path, infos["class"], infos["props"])

    if res["DeviceType"] == nm_device_type[1]:
        # Ethernet
        e = {"class": "org.freedesktop.NetworkManager.Device.Wired", "props": [
                {"name": "HwAddress", "map": identity, "sub": None},
                {"name": "Carrier", "map": identity, "sub": None}
        ]}
        eth_props = _get_properties(NM_BUS_NAME, dev_path, e["class"], e["props"])
        for k, v in eth_props.items():
            res[k] = v
    elif res["DeviceType"] == nm_device_type[2]:
        # Wifi
        w = {"class": "org.freedesktop.NetworkManager.Device.Wireless", "props": [
                {"name": "HwAddress", "map": identity, "sub": None},
                {"name": "Mode", "map": lambda x: nm_80211_mode[x], "sub": None},
                {"name": "Bitrate", "map": identity, "sub": None}
        ]}
        wifi_props = _get_properties(NM_BUS_NAME, dev_path, w["class"], w["props"])
        for k, v in wifi_props.items():
            res[k] = v

    # active connection configuration
    if "ActiveConnection" in res.keys() and "Connection" in res["ActiveConnection"].keys() and \
            res["ActiveConnection"]["Connection"] is not None:
        proxy = bus.get_object(NM_BUS_NAME, res["ActiveConnection"]["Connection"])
        settings_connection = dbus.Interface(proxy, "org.freedesktop.NetworkManager.Settings.Connection")
        res["ActiveConnection"]["Connection"] = settings_connection.GetSettings()

    return res


@catch_dbus_conn_err
def get_nm_state():
    return nm_state[manager.state()]


@catch_dbus_conn_err
def scan_aps(interface):
    """trigger AP scan on a given interface"""
    dev_path = manager.GetDeviceByIpIface(interface)

    proxy = bus.get_object(NM_BUS_NAME, dev_path)
    iface = dbus.Interface(proxy, "org.freedesktop.NetworkManager.Device.Wireless")

    try:
        iface.RequestScan({})
    except dbus.exceptions.DBusException as e:
        # ignore when device is not active
        if e.get_dbus_name() == "org.freedesktop.NetworkManager.Device.NotAllowed":
            logger.error("Failed to request scan: {}".format(e))
        else:
            raise e


@catch_dbus_conn_err
def get_aps(interface):
    """retrieve access points of an interface"""
    dev_path = manager.GetDeviceByIpIface(interface)

    ap_props = {"class": "org.freedesktop.NetworkManager.AccessPoint", "props": [
            {"name": "HwAddress", "map": identity, "sub": None},
            {"name": "LastSeen", "map": identity, "sub": None},
            {"name": "Strength", "map": identity, "sub": None},
            {"name": "Mode", "map": lambda x: nm_80211_mode[x], "sub": None},
            {"name": "Ssid", "map": lambda x: bytearray(x).decode(), "sub": None},
            {"name": "Frequency", "map": identity, "sub": None}
    ]}

    proxy = bus.get_object(NM_BUS_NAME, dev_path)
    iface = dbus.Interface(proxy, "org.freedesktop.NetworkManager.Device.Wireless")
    ap_refs = iface.GetAllAccessPoints()
    res = [_get_properties(NM_BUS_NAME, path, ap_props["class"], ap_props["props"]) for path in ap_refs]

    return res


@catch_dbus_conn_err
def get_conn_by_uuid(conn_uuid):
    """Get connection settings by UUID
    :returns (path, setting) of the connection or (None, None) if it doesn't exist
    """
    for path in settings.ListConnections():
        proxy = bus.get_object(NM_BUS_NAME, path)
        settings_connection = dbus.Interface(proxy, "org.freedesktop.NetworkManager.Settings.Connection")
        config = settings_connection.GetSettings()
        if config["connection"]["uuid"] == conn_uuid:
            return path, settings_connection

    return None, None


def detect_device(dev_type, name=None):
    dev = get_device_by_type(dev_type)
    if len(dev) == 0:
        logger.error("No {} device detected!".format(dev_type))
        sys.exit(2)

    if name is None:
        if len(dev) > 1:
            logger.warning("{} {} devices detected, using {}".format(len(dev), dev_type, ))
        else:
            logger.info("Managing {} device {}".format(dev_type, dev[0]["Interface"]))
        return str(dev[0]["Interface"])
    else:
        match = [d["Interface"] for d in dev if d["Interface"] == name]
        return None if len(match) == 0 else match[0]


@decorate_all_methods(catch_dbus_conn_err)
class SimpleNetworkManager(object):
    """Manages a single wifi and ethernet interface and updates a single, hard-coded connection for each of these"""
    def __init__(self, eth_iface=None,
                 wifi_iface=None,
                 wifi_ap_iface=None,
                 eth_conn_uuid="d6a22aa3-c789-438e-97d9-d73fb9b52499",
                 wifi_conn_uuid="40b90a3c-8c68-42c4-a3b9-0f127266c7ac",
                 wifi_ap_conn_uuid="77c902ee-ebef-4ae6-beb3-eb49f945bfb5"):

        self.eth = eth_iface if eth_iface is not None else detect_device("ETHERNET")

        # if there are multiple wifi interfaces, use wlan-int as wifi ap
        # if not, remove wifi ap connection (if it existed) and use wlan0 in STA mode
        num_wifi = get_device_by_type("WIFI")
        if len(num_wifi) < 2:
            # remove wifi ap connection if exists
            _, conn_settings = get_conn_by_uuid(wifi_ap_conn_uuid)
            if conn_settings is not None:
                conn_settings.Delete()

            self.wifi_ap = None
            self.wifi = wifi_iface if wifi_iface is not None else detect_device("WIFI", "wlan0")
        else:
            self.wifi = wifi_iface if wifi_iface is not None else self.detect_wifi_device()
            self.wifi_ap = wifi_ap_iface if wifi_ap_iface is not None else detect_device("WIFI", "wlan0")

        self.eth_conn_uuid = eth_conn_uuid
        self.wifi_conn_uuid = wifi_conn_uuid
        self.wifi_ap_conn_uuid = wifi_ap_conn_uuid

        logger.info("Wifi interface {}".format(self.wifi))
        logger.info("Wifi AP interface {}".format(self.wifi_ap))
        logger.info("Ethernet interface {}".format(self.eth))

        self.uuids = {self.eth: eth_conn_uuid,
                      self.wifi: wifi_conn_uuid,
                      self.wifi_ap: wifi_ap_conn_uuid}

        self._aps = []

    def detect_wifi_device(self):
        """Helper method to get Wifi device which is not wlan0 (for AP mode)"""
        for i in range(1, 7):
            d = detect_device("WIFI", "wlan{}".format(i))
            if d is not None:
                return d
        return None

    def update_connection(self, iface, new_conn_settings, override=True):
        if iface not in self.uuids.keys():
            raise Exception("Device {} not managed".format(iface))

        dev_path = manager.GetDeviceByIpIface(iface)
        specific_obj = dbus.ObjectPath("/")

        conn_name = "{}-default".format(iface)
        if iface == self.wifi_ap:
            conn_name = "wifi-ap-default"

        new_conn_settings["connection"]["uuid"] = self.uuids[iface]
        new_conn_settings["connection"]["id"] = conn_name

        conn_path, conn_settings = get_conn_by_uuid(self.uuids[iface])
        if conn_settings is None:
            # create a new one for this device
            logger.info(new_conn_settings)

            manager.AddAndActivateConnection(new_conn_settings, dev_path, specific_obj)
        else:
            logger.info("Updating connection for {}: {}".format(iface, new_conn_settings))
            if override:
                conn_settings.Update(new_conn_settings)

            manager.ActivateConnection(conn_path, dev_path, specific_obj)

    def disable(self, iface, delete=True):
        """Delete managed connection from device and disconnect"""
        if iface is None:
            return

        if iface not in self.uuids.keys():
            raise Exception("Device {} not managed".format(iface))

        conn_path, conn_settings = get_conn_by_uuid(self.uuids[iface])
        if conn_settings is not None and delete:
            conn_settings.Delete()

        dev_path = manager.GetDeviceByIpIface(iface)

        proxy = bus.get_object(NM_BUS_NAME, dev_path)
        iface = dbus.Interface(proxy, "org.freedesktop.NetworkManager.Device")
        try:
            iface.Disconnect()
        except dbus.exceptions.DBusException as e:
            # ignore when device is not active
            if e.get_dbus_name() != "org.freedesktop.NetworkManager.Device.NotActive":
                raise e

    def scan_aps(self):
        return scan_aps(self.wifi)

    def get_wifi_aps(self):
        aps = get_aps(self.wifi)
        if aps is not None and len(aps) > 0:
            self._aps = aps

        return self._aps

    def enable_wifi_ap(self, ap_settings=None):
        if ap_settings is None:
            ap_settings = DEFAULT_WIFI_HOTSPOT_SETTINGS

        if self.wifi_ap is None:
            logger.warning("No WiFi AP interface")
            return

        dev_path = manager.GetDeviceByIpIface(self.wifi_ap)

        conn_path, conn = get_conn_by_uuid(self.wifi_ap_conn_uuid)

        ap_settings["connection"]["uuid"] = self.uuids[self.wifi_ap]
        ap_settings["connection"]["id"] = "wifi-ap-default".format(self.wifi_ap)

        if conn_path is None:
            conn_path = settings.AddConnection(ap_settings)
        else:
            conn.Update(ap_settings)

        logger.info("Activating Wifi AP for device {}".format(self.wifi_ap))
        manager.ActivateConnection(conn_path, dev_path, "/")

    def get_device_states(self):
        states = [get_device_state(self.wifi), get_device_state(self.eth)]

        for s in states:
            if "ActiveConnection" in s and s["ActiveConnection"] is not None and "Connection" in s["ActiveConnection"]:
                s["ActiveConnection"]["Connection"] = self._map_conn_settings_out(s["ActiveConnection"]["Connection"])

        return states

    def get_connection(self, iface):
        if iface not in self.uuids.keys():
            raise Exception("Device {} not managed".format(iface))

        conn_path, settings_conn = get_conn_by_uuid(self.uuids[iface])
        if settings_conn is not None:
            return self._map_conn_settings_out(settings_conn.GetSettings())

        return None

    def init_connections(self, force=False):
        """Remove all except the managed connection IDs, init ethernet with DHCP and WiFi with Hotspot connection
        :param force: when True force reset of all network settings (default False)
        """
        connections = []
        for path in settings.ListConnections():
            proxy = bus.get_object(NM_BUS_NAME, path)
            settings_connection = dbus.Interface(proxy, "org.freedesktop.NetworkManager.Settings.Connection")
            config = settings_connection.GetSettings()
            connections.append((config["connection"]["uuid"], settings_connection))

        if not force:
            has_wifi_conn = any(uuid == self.wifi_conn_uuid for uuid, c in connections)
            has_wifi_ap_conn = any(uuid == self.wifi_ap_conn_uuid for uuid, c in connections)
            has_eth_conn = any(uuid == self.eth_conn_uuid for uuid, c in connections)

            if has_wifi_conn:
                self.update_connection(self.wifi, self.get_connection(self.wifi), override=False)
            if has_wifi_ap_conn and self.wifi_ap is not None:
                self.update_connection(self.wifi_ap, self.get_connection(self.wifi_ap), override=False)

            if has_wifi_conn and has_eth_conn:
                logger.info("Managed connections already present")
                return

        # otherwise, delete other connections and set up defaults
        logger.info("Deleting connections we don't manage")
        for uuid, conn in connections:
            if not uuid in self.uuids.values():
                conn.Delete()

        logger.info("Scanning for WiFi Networks")
        try:
            self.scan_aps()
        except Exception as e:
            logger.error("Failed to trigger AP scan: {}".format(e))

        logger.info("Setting up default DHCP Ethernet connection")
        self.update_connection(self.eth, DEFAULT_ETHERNET_SETTINGS)
        self.enable_wifi_ap()

    def _map_conn_settings_out(self, settings):
        if settings is None:
            return None

        if "802-11-wireless" in settings.keys() and "ssid" in settings["802-11-wireless"].keys():
            settings["802-11-wireless"]["ssid"] = "".join(chr(b) for b in settings["802-11-wireless"]["ssid"])

        return settings


if __name__ == "__main__":
    LOG_FORMAT = "%(asctime)-15s %(name)s[%(threadName)s] %(levelname)-7s %(message)s"
    logging.basicConfig(level=logging.INFO, format=LOG_FORMAT)

    #ETH_UUID = "f44e758d-7989-49b9-9b7d-68599b3aa38e"
    #WIFI_UUID = "95490b65-fd30-4bea-b2a3-bfa67b1fa80e"

    #sm = SimpleNetworkManager("enp0s8", "enp0s3", ETH_UUID, WIFI_UUID)

    # logger.info(json.dumps(sm.get_device_states()))
    # logger.info("NetworkManager state is: '%s'" % get_nm_state())
    # sm.update_connection(sm.eth, eth_dhcp)
    # logger.info(json.dumps(sm.get_connection(sm.eth)))
    # logger.info(json.dumps(sm.get_wifi_aps()))
    # sm.disable(sm.wifi)

    s1 = {
        "802-3-ethernet": {
            "auto-negotiate": 0,
            "mac-address": [
                184,
                39,
                235,
                126,
                214,
                59
            ],
            "mac-address-blacklist": [],
            "s390-options": {}
        },
        "connection": {
            "id": "eth0-default",
            "permissions": [],
            "timestamp": 1567154831,
            "type": "802-3-ethernet",
            "uuid": "d6a22aa3-c789-438e-97d9-d73fb9b52499"
        },
        "ipv4": {
            "address-data": [],
            "addresses": [],
            "dns": [],
            "dns-search": [],
            "method": "auto",
            "route-data": [],
            "routes": []
        },
        "ipv6": {
            "address-data": [],
            "addresses": [],
            "dns": [],
            "dns-search": [],
            "method": "ignore",
            "route-data": [],
            "routes": []
        },
        "proxy": {}
    }
    print(transform_settings(s1))
